import re
import os
import socket
import asyncio
from asyncio import shield
import math
import errno
from struct import unpack
from functools import reduce
import sys
import time
import datetime


from basic_logtools.filelog import LogFile

from networktools.time import timestamp, now
from networktools.library import my_random_string
'''
###############################################################################
CLASE ERYO  __authors__ = " David Pineda"
###############################################################################
Conexion TCP/IP con un determinado receptor.
'''
from .eryo_settings import (ERYO_FIELD_NAMES, ERYO_REC, ERYO_TABLE_NAMES,
                            BUFF_MAX_SIZE, TABLES_SIZES, ERYO_LIST, SYNC_MARKER_VALUE, BLOCKS_FLAGS,
                            ETN_reverse, TABLES_BY_FLAG, CONS, MSG_ID, signal_check)

from networktools.colorprint import rprint, gprint, bprint

from dataprotocols import BaseProtocol

"""
Check verification

int eryo_trailer_decode(char *buf, int start, eryo_trailer_t *ptr)
{
  int index=start;
  DECODE_BITS(16, buf, index, ptr->checksum);
  assert(index - start==sizeof(eryo_trailer_t));
  return index;
}

inidice = partida
decodificar 2 bytes (16 bits) a variable checksu,
compara,  la diferencia (index-start) con tamaño de eryo_trailer_y


"""


def convert_data(key, table_name, data):
    size_table = ERYO_REC[table_name]
    field_names = ERYO_FIELD_NAMES[key]
    bytes_table = TABLES_SIZES[table_name]
    try:
        unpacked = unpack(size_table, data)
    except Exception as e:
        print("No unpack on converta data", e)
        raise e
    table_dict = dict(zip(field_names, unpacked))
    return table_dict


def checksum(data):
    check_data = list(data)
    suma = sum(check_data[:-2])
    return suma


class Eryo(BaseProtocol):
    """ Class to connect to tcp port and parse ERYO messages """
    tipo = "ERYO"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.msg_0 = None
        self.msg_bytes = None
        self.rec_dict = {}
        self.batt_status = 0
        self.tables = ERYO_TABLE_NAMES

    async def get_records(self):
        """
        Obtener bytes y traducir a un msg ERYO

        """
        idc = self.select_idc
        reader = self.clients[idc]['reader']
        writer = self.clients[idc]['writer']
        sync_marker_first = SYNC_MARKER_VALUE.get('first')
        sync_marker_second = SYNC_MARKER_VALUE.get('second')
        t0 = time.time()
        try:
            if writer.transport.is_closing():
                self.logger.error(f"La conexión se cerró {writer}")
                self.status = False
                raise Exception("Conexión perdida")
            try:
                done = False
                counter = 1
                data = b''
                byte_a, byte_b = (0, 0)
                suma_bytes = 0
                msg = {}
                position = 0
                while not done:
                    byte_a = await shield(self.readbytes(reader, 1))
                    while byte_a != sync_marker_first:
                        byte_a = await shield(self.readbytes(reader, 1))
                    byte_b = await shield(self.readbytes(reader, 1))
                    if byte_b == sync_marker_second:
                        # read next 3 bytes
                        resto_header = await shield(self.readbytes(reader, 3))
                        suma_bytes += 5
                        header = byte_a+byte_b+resto_header
                        data += header
                        data_dict = convert_data(1, "ERYO_HEADER", header)
                        data_dict['MESSAGE_ID'] = MSG_ID.get(
                            data_dict.get("MESSAGE_ID"))
                        msg["ERYO_HEADER"] = data_dict
                        # POSITION_BLOCK
                        byte_count = data_dict.get("BYTE_COUNT")
                        all_bytes = await shield(self.readbytes(reader, byte_count-5))
                        data += all_bytes
                        key_2 = "POSITION_BLOCK"
                        PB_bytes = TABLES_SIZES.get(key_2)
                        suma_bytes += PB_bytes
                        PB_data = all_bytes[position:position+PB_bytes]
                        position += PB_bytes
                        PB_dict = convert_data(2, key_2, PB_data)
                        SITE_ID = PB_dict["SITE_ID"].decode("utf8")
                        PB_dict["SITE_ID"] = re.sub("\x00+$", "", SITE_ID)
                        SOLUTION_ID = PB_dict["SOLUTION_ID"].decode("utf8")
                        PB_dict["SOLUTION_ID"] = re.sub(
                            "\x00+$", "", SOLUTION_ID)
                        msg["POSITION_BLOCK"] = PB_dict
                        FLAGS = PB_dict.get('FLAGS')
                        bloques = list(filter(
                            lambda e: e[1] > 0,
                            [(block, FLAGS & value) for block, value in
                             BLOCKS_FLAGS.items()]))
                        for block_name, value in bloques:
                            if block_name == 'SAT_INFO':
                                HDR = "SAT_HDR"
                                INFO = "SAT_INFO_BLOCK"
                                nbytes = TABLES_SIZES.get(HDR)
                                suma_bytes += nbytes
                                data_block = all_bytes[position:position+nbytes]
                                position += nbytes
                                n = ETN_reverse.get(HDR)
                                data_dict = convert_data(n, HDR, data_block)
                                msg["SAT_INFO"] = {HDR: data_dict}
                                msg["SAT_INFO"]["DATA"] = []
                                NRO_SATS = data_dict.get('SAT_BLOCK_COUNT')
                                key = TABLES_BY_FLAG.get("SAT_INFO")[1]
                                for i in range(NRO_SATS):
                                    nbytes = TABLES_SIZES.get(INFO)
                                    suma_bytes += nbytes
                                    data_block = all_bytes[position:position+nbytes]
                                    position += nbytes
                                    n = ETN_reverse.get(INFO)
                                    data_dict = convert_data(
                                        n, INFO, data_block)
                                    key = "CONSTELLATION"
                                    data_dict[key] = CONS.get(
                                        data_dict.get(key))
                                    key = "SIGNAL_FLAGS"
                                    data_dict["SIGNALS"] = signal_check(
                                        data_dict.get(key))
                                    msg["SAT_INFO"]["DATA"].append(data_dict)
                            else:
                                for key in TABLES_BY_FLAG.get(block_name):
                                    nbytes = TABLES_SIZES.get(key)
                                    suma_bytes += nbytes
                                    data_block = all_bytes[position:position+nbytes]
                                    position += nbytes
                                    n = ETN_reverse.get(key)
                                    data_dict = convert_data(
                                        n, key, data_block)
                                    msg[key] = data_dict
                        # CHECKSUM
                        key = "ERYO_TRAILER"
                        nbytes = TABLES_SIZES.get(key)
                        suma_bytes += nbytes
                        checksum_bytes = all_bytes[position:position+nbytes]
                        n = ETN_reverse.get(key)
                        # checksum_bytes = await reader.readexactly(nbytes)
                        data_dict = convert_data(n, key, checksum_bytes)
                        msg[key] = data_dict
                        checksum_msg = data_dict.get("CHECKSUM")
                        chsum = checksum(data)
                        done = checksum_msg == chsum
                return done, msg
            except asyncio.TimeoutError as te:
                print(f"Timeout error on Eryo {self.station}, error {te}")
                self.logger.exception(
                    f"Get message records TimeoutError, {te}")
                return False, {}
            except asyncio.IncompleteReadError as ir:
                print(
                    f"Incomplete read on get records header, ERYO, station {self.station}")
                self.logger.exception(
                    f"Station {self.station}, Tiempo fuera al no poder leer en readbytes {n} bytes, {ir}")
                if self.raise_incompleteread:
                    raise ir
                return False, {}
            except socket.timeout as timeout:
                self.logger.warning(
                    f"La conexión se cerró, socket timeout {timeout}")
                self.on_blocking()
                self.logger.exception()
                self.status = False
                await asyncio.sleep(.1)
                return False, {}
            except Exception as ex:
                self.logger.error(
                    f"La conexión se cerró, error en comunicación {ex}")
                self.close(idc)
                self.on_blocking()
                info = f"Error {ex} en estación {self.station}. Encabezado con data {data}"
                self.logger.info(info)
                self.logger.error(ex)
                self.status = False
                await asyncio.sleep(.1)
                return False, {}
        except socket.timeout as timeout:
            self.logger.warning(
                f"La conexión se cerró, socket timeout {timeout}")
            self.on_blocking()
            self.logger.exception()
            self.status = False
            await asyncio.sleep(.1)
            return False, {}
        except Exception as ex:
            self.logger.error(
                f"La conexión se cerró, error en comunicación {ex}")
            self.close(idc)
            self.on_blocking()
            info = f"Error {ex} en estación {self.station}. Encabezado con data {data}"
            self.logger.info(info)
            self.logger.error(ex)
            self.status = False
            await asyncio.sleep(.1)
            return False, {}

    async def get_message_header(self, idc):
        self.select_idc = idc



from dataprotocols.gsof import Gsof
import asyncio
import functools
import os
import signal
import click
import json
import time
from networktools.ip import validURL


async def run_test(loop, code, host, port, limit=2000, show=True):
    ginput = dict(host=host,
                  port=port,
                  code=code,
                  timeout=5,
                  sock=None,
                  loop=loop)
    print("Inicializando GSOF con:")
    [print(k, "->", v) for k, v in ginput.items()]
    gsof_test = "gsof_test.json"
    limit_counter = limit
    counter = 0
    print("Conexion a->", ginput)
    gsof = Gsof(**ginput)
    print("GSOF object", gsof)
    idc = await gsof.connect()
    print("IDC->", idc)
    with open(gsof_test, "a+") as gsof_file:
        gsof_file.write("[")
        while counter <= limit_counter:
            try:
                print("Heart beat")
                heartbeat = await gsof.heart_beat(idc)
                print("Heartbeat", heartbeat)
                if heartbeat:
                    print("MSG recv>")
                    await gsof.get_message_header(idc)
                    print("Header msg->", gsof.msg_bytes)
                    done, msg = await gsof.get_records()
                    if done:
                        if show:
                            print("Gsof %s" % code)
                            [print(k, "->", v) for k, v in msg.items()]
                        json.dump(msg, gsof_file, indent=2)
                        gsof_file.write(",\n")
                else:
                    await gsof.close(idc)
                    idc = await gsof.connect()
            except Exception as ex:
                gsof_file.write("]")
                await gsof.close(idc)
                print("Error %s" % ex)
                loop.close()
                raise ex
            except KeyboardInterrupt as ke:
                gsof_file.write("]")
                await gsof.close(idc)
                loop.close()
                print(ke)
                raise ke
            counter += 1
        gsof_file.write("]")
    await gsof.close(idc)


@click.command()
@click.option("--code", default="CODE", help="Código de estación")
@click.option("--host", default="AQUI-VA-LA-URL", help="URL de estación")
@click.option("--port", default=9080, help="Nro de puerto")
@click.option("--limit", default=2000, help="Nro de mensajes")
@click.option("--show/--no-show", default=True, show_default=True,
              type=bool, help="Mostrar cada mensaje")
def run_gsof(code, host, port, limit, show):
    loop = asyncio.get_event_loop()
    try:
        if validURL(host):
            loop.run_until_complete(
                run_test(loop, code, host, port, limit, show))
        else:
            print(
                "Ingresa datos reales para obtener resultados: gsof --help \n Revisa tus parámetros nuevamente")
    except Exception as ex:
        print("Exception %s" % ex)
        loop.call_soon_threadsafe(loop.stop)
        print(ex)
    except KeyboardInterrupt as ke:
        print("Exception %s" % ke)
        loop.call_soon_threadsafe(loop.stop)
        print(ke)
    finally:
        loop.close()


if __name__ == '__main__':
    run_gsof()



from struct import unpack
import socket
from asyncio import shield
import asyncio

from networktools.time import timestamp

from networktools.colorprint import rprint, bprint

from dataprotocols import BaseProtocol

'''
###############################################################################
CLASE GSOF  __authors__ = "Henry T. Berglund ----> David Pineda"
###############################################################################
Conexion TCP/IP con un determinado receptor.
Updated to python3.5 by David Pineda @pineiden dpineda@csn.uchile.cl

'''

try:
    from .gsof_settings import gsof_field_names, gsof_rec, gsof_table_names
except Exception:
    from gsof_settings import gsof_field_names, gsof_rec, gsof_table_names


def checksum256(data_check):
    """Calculate checksum"""
    # ord: string -> unicode int
    # map(a,b): opera a en b
    # list entrega lista o arrays
    # formula: 	(Status + type + length + data bytes) modulo 256
    data_bytes = data_check.get("DATA_BYTES")
    checksum = data_check.get("CHECKSUM")
    chk256 = sum(data_bytes) % 256
    return checksum == chk256


class Gsof(BaseProtocol):
    """ Class to connect to tcp port and parse GSOF messages """
    tipo = "GSOF"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.msg_0 = None
        self.msg_bytes = None
        self.rec_dict = {}
        self.batt_status = 0
        self.tables = gsof_table_names
        # manage asyncronous clients

    async def get_message_header(self, idc):
        # receive data from source
        # max amount of data
        # N ammount of bites
        self.data_check = {}
        reader = self.clients[idc]['reader']
        writer = self.clients[idc]['writer']
        self.timestamp = timestamp()
        BytesData = 7
        bufsize = BytesData
        msg_field_names = ('STX', 'STATUS', 'TYPE', 'LENGTH',
                           'T_NUM', 'PAGE_INDEX', 'MAX_PAGE_INDEX')

        try:
            if writer.transport.is_closing():
                self.logger.error(f"La conexión se cerró {writer}")
                self.status = False
                raise Exception(
                    f"Conexión perdida para idc {idc} station {self.station}")
            data = None

            try:
                data = await shield(self.readbytes(reader, bufsize))
                self.msg_dict = dict(zip(
                    msg_field_names,
                    unpack('>' + str(bufsize) + 'B', data)))
                campos_unpack = unpack('>%dB' % bufsize, data)
                self.msg_dict = dict(zip(msg_field_names, campos_unpack))
                status = self.msg_dict['STATUS']
                bin_status = bin(status)
                batt_status = bin_status[3]
                self.msg_dict["LOW_BATTERY"] = bool(int(batt_status))
                self.msg_bytes = await shield(self.readbytes(reader, self.msg_dict['LENGTH']-3))
                self.msg_0 = self.msg_bytes
                n = 2
                check_etx = await shield(self.readbytes(reader, n))
                (checksum, etx) = unpack('>2B', check_etx)
                self.data_check["CHECKSUM"] = checksum
                self.data_check["DATA_BYTES"] = list(self.msg_bytes+data[1:])
                self.checksum = checksum256(self.data_check)
                self.msg_dict["CHECKED"] = self.checksum
            except asyncio.TimeoutError as te:
                self.logger.exception("Get message TimeoutError, %s" % te)
                return False
            except asyncio.IncompleteReadError as ir:
                print(f"Incomplete read on set header {reader}, {n} bytes")
                self.logger.exception(
                    f"Station {self.station}, Tiempo fuera al no poder leer en readbytes {n} bytes, {ir}")
                return False
            except socket.timeout as timeout:
                self.logger.exception(
                    f"La conexión se cerró, socket timeout {timeout}, STATION {self.station}")
                self.on_blocking()
                self.logger.exception("Esperando mucho tiempo")
                self.status = False
                return False
            except Exception as ex:
                print("Escepcion as ex: %s" % ex)
                self.logger.exception(
                    f"La conexión se cerró, error en comunicación {ex}")
                await self.close(idc)
                self.on_blocking()
                info = f"Error {ex} en estación {self.station}. Encabezado con data {data}"
                self.logger.info(info)
                self.logger.error(ex)
                self.status = False
                return False
            return True

        except asyncio.IncompleteReadError as e:
            self.logger.exception(f"Error de mensaje incompleto {e}")
            self.logger.error(
                f"Error de mensaje esperado en n bytes {e.expected}")
            self.logger.error(f"Error de mensaje recibido {e.partial}")
            raise e

        except socket.timeout:
            self.close()
            self.sock = self.create_socket(None)
            self.on_blocking()
            raise socket.timeout

        except socket.error as ex:
            self.close()
            self.logger.error(ex)
            self.sock = self.create_socket(None)
            self.on_blocking()
            raise ex

    async def get_records(self):
        msg = {}
        while self.msg_bytes:
            # READ THE FIRST TWO BYTES FROM RECORD HEADER
            try:
                record_type, record_length = unpack('>2B', self.msg_bytes[0:2])
            except Exception as ex:
                self.status = False
                qex = f"Error {ex} en estación {self.station}, tabla Header"
                self.logger.debug(qex)
                return False, {}
            self.msg_bytes = self.msg_bytes[2:]
            try:
                msg_table = self.select_record(record_type, record_length)
                msg.update(msg_table)
            except Exception as ex:
                self.status = False
                self.logger.error(ex)
                self.logger.error("Desconectando debido a error en mensaje")
                self.logger.error(
                    f"Reconectando debido a error en mensaje, luego de desconexión, {ex}")
                self.on_blocking()
                return False, {}
        msg["HEADER"] = self.msg_dict
        done = self.checksum
        return done, msg

    def select_record(self, record_type, record_length):
        field_names = gsof_field_names
        rec_input = gsof_rec
        msg = []
        table_name = self.tables.get(record_type)
        timestamp = self.timestamp
        try:
            rec_field_names = field_names.get(record_type)
            if record_type not in [13, 14, 33, 48] and record_type in field_names.keys():
                # print("Base recibida")
                try:
                    rec_values = unpack(
                        rec_input[record_type], self.msg_bytes[:record_length])
                    msg = dict(list(zip(rec_field_names, rec_values)))
                    msg['TIMESTAMP'] = timestamp
                    self.rec_dict.update({table_name: msg})
                    self.msg_bytes = self.msg_bytes[record_length:]
                except Exception as ex:
                    self.status = False
                    self.logger.exception(f"Error al hacer unpack {ex}")
                    raise ex

            elif record_type in [13, 14, 33, 48] and record_type in field_names.keys():
                L = len(self.msg_bytes)-1
                bNUM = self.msg_bytes[0:1]
                try:
                    code = rec_input[record_type][0]
                    NUM_OF_SVS = unpack(code, bNUM)  # char-> pass to int
                except Exception as ex:
                    self.status = False
                    qex = "Error %s en estación %s, tabla %s" % (
                        ex, self.station, record_type)
                    self.logger.exception(qex)
                    raise ex
                u = int(L/NUM_OF_SVS[0])
                # Acortar el NSVS quitando primer byte
                self.msg_bytes = self.msg_bytes[1:]
                # for field in range(len(rec_field_names)):
                #    self.rec_dict[rec_field_names[field]] = []
                for sat in range(NUM_OF_SVS[0]):
                    try:
                        code = rec_input[record_type][1]
                        rec_values = unpack(
                            code, self.msg_bytes[0:u])  # why [0:10]
                        msg_k = dict(list(zip(rec_field_names, rec_values)))
                        msg_k['TIMESTAMP'] = timestamp
                        msg.append(msg_k)
                    except Exception as ex:
                        self.status = False
                        self.logger.debug(ex)
                        raise ex
                    self.msg_bytes = self.msg_bytes[u:]
                    tname = table_name+"_"+str(sat)
                    self.rec_dict.update({table_name: msg})
            else:
                """Unknown record type? Skip it for now!"""
                # print record_type
                msg = self.msg_bytes[record_length + 2:]
                self.msg_bytes = self.msg_bytes[record_length + 2:]
        except Exception as ex:
            self.status = False
            raise ex
        return {table_name: msg}



from networktools.time import timestamp

# uint8_t
PROTOCOL_VERSION = 2.3
PV_INT = int(PROTOCOL_VERSION*10)


ERYO_TABLE_REC = dict(
    ERYO_HEADER=">2bHb",
    POSITION_BLOCK=">bHIHH8s8sbb4db",
    POS_VAR_BLOCK=">7f",
    TROPO_VAR_BLOCK=">4f",
    SAT_HDR=">Hf",
    SAT_INFO_BLOCK=">3BbH",
    NEU_DISP_HDR=">3dH",
    NEU_DISP_BLOCK=">3f",
    VELOCITY_HDR=">H",
    VELOCITY_BLOCK=">3f",
    NEU_VAR_BLOCK=">6f",
    VEL_VAR_BLOCK=">6f",
    ERYO_TRAILER=">H",
)

ERYO_TABLE_NAMES = {
    1: "ERYO_HEADER",
    2: "POSITION_BLOCK",
    3: "POS_VAR_BLOCK",
    4: "TROPO_VAR_BLOCK",
    5: "SAT_HDR",
    6: "SAT_INFO_BLOCK",
    7: "NEU_DISP_HDR",
    8: "NEU_DISP_BLOCK",
    9: "VELOCITY_HDR",
    10: "VELOCITY_BLOCK",
    11: "NEU_VAR_BLOCK",
    12: "VEL_VAR_BLOCK",
    13: "ERYO_TRAILER",
}

ETN_reverse = {value: key for key, value in ERYO_TABLE_NAMES.items()}


ERYO_FIELD_NAMES = {
    1: ("SYNC_MARKER_1", "SYNC_MARKER_2", "BYTE_COUNT", "MESSAGE_ID"),
    2: ("ERYO_VERSION",
        "GPS_WEEK",
        "GPS_MILLISECONDS",
        "SITE_INDEX",
        "SITES_THIS_EPOCH",
        "SITE_ID",
        "SOLUTION_ID",
        "ECEF_FRAME",
        "AMBIGUITY_RES",
        "XYZ_ZENITH_X",  "XYZ_ZENITH_Y", "XYZ_ZENITH_Z",
        "TROPO_ZENITH_DELAY",
        "FLAGS"),
    3: ("SCALE_FACTOR", "X_VAR", "Y_VAR", "Z_VAR", "YX_COV", "YZ_COV", "ZX_COV"),
    4: ("T_VARIANCE", "TX_COV", "TY_COV", "TZ_COV"),
    5: ("SAT_BLOCK_COUNT", "GDOP"),
    6: ("CONSTELLATION", "SAT_NUMBER", "SIGNAL_FLAGS",
        "PRN_ELEVATION", "PRN_AZIMUTH"),
    7: ("X", "Y", "Z"),
    8: ("N", "E", "U"),
    9: ("VELOCITY_BLOCK_COUNT"),
    10: ("VX", "VY", "VZ"),
    11: ("N_POS_VAR", "E_POS_VAR", "U_POS_VAR",
         "EN_POS_COV", "EU_POS_COV", "UN_POS_COV"),
    12: ("N_VEL_VAR", "E_VEL_VAR", "U_VEL_VAR",
         "EN_VEL_COVAR", "EU_VEL_COVAR", "UN_VEL_COVAR"),
    13: ("CHECKSUM",),
}

nros = [1 << i for i in range(8)]
MSG_ID = dict(zip(nros, ["NEU", "VEL", "RES2", "RES3", "RES4", "NET_PPP", "RES6",
                         "RT_POST"]))


# para comparación binaria FLAG & valor->010101
BLOCKS_FLAGS = dict(
    POS_VAR=1 << 0,
    TROPO_VAR=1 << 1,
    SAT_INFO=1 << 2,
    NEU_DISP=1 << 3,
    VELOCITY=1 << 4,
    NEU_VAR=1 << 5,
    VEL_VAR=1 << 6,
    RES=1 << 7,
)

TABLES_BY_FLAG = dict(
    POS_VAR=("POS_VAR_BLOCK",),
    TROPO_VAR=("TROPO_VAR_BLOCK",),
    SAT_INFO=("SAT_HDR", "SAT_INFO_BLOCK",),
    NEU_DISP=("NEU_DISP_HDR", "NEU_DISP_BLOCK"),
    VELOCITY=("VELOCITY_HDR", "VELOCITY_BLOCK"),
    NEU_VAR=("NEU_VAR_BLOCK",),
    VEL_VAR=("VEL_VAR_BLOCK",),
    RES=(),
)


lista = [(k, v) for k, v in ERYO_TABLE_NAMES.items()]

ERYO_LIST = sorted(lista, key=lambda e: e[0])


ERYO_REC = dict(
    ERYO_HEADER=">BBHB",  # 1,1,2,1
    POSITION_BLOCK=">BHIHH8s8sBB4dB",
    POS_VAR_BLOCK=">7f",
    TROPO_VAR_BLOCK=">4f",
    SAT_HDR=">Hf",
    SAT_INFO_BLOCK=">3BbH",
    NEU_DISP_HDR=">3dH",
    NEU_DISP_BLOCK=">3f",
    VELOCITY_HDR=">H",
    VELOCITY_BLOCK=">3f",
    NEU_VAR_BLOCK=">6f",
    VEL_VAR_BLOCK=">6f",
    ERYO_TRAILER=">H")

TABLES_SIZES = dict(
    ERYO_HEADER=5,
    POSITION_BLOCK=62,
    POS_VAR_BLOCK=28,
    TROPO_VAR_BLOCK=16,
    SAT_HDR=6,
    SAT_INFO_BLOCK=6,
    NEU_DISP_HDR=26,
    NEU_DISP_BLOCK=12,
    VELOCITY_HDR=2,
    VELOCITY_BLOCK=12,
    NEU_VAR_BLOCK=24,
    VEL_VAR_BLOCK=24,
    ERYO_TRAILER=2,
)


MSG_MAX_SIZE = sum(TABLES_SIZES.values())  # bytes

BUFF_MAX_SIZE = MSG_MAX_SIZE


def logfile(station):
    return './logs/gsof_%s_%s.log' % (station, str(timestamp())[:-7])


SYNC_MARKER_VALUE = {
    "first": b"\x9C",
    "second": b"\xA5",
}

"""
checksum_t eryo_calculateChecksum(
         const char * const buffer,
         const int buff_size)
{
  checksum_t sum=(checksum_t) 0;
  uint8_t *ptr, *last;
  ptr=(uint8_t *) &buffer[0];1
  last=(uint8_t *) &buffer[buff_size - 1];
  while (ptr <=last)
  {
    sum +=(checksum_t) *ptr++;
  }
  return sum;
}
"""


def eryo_checksum(buffer, buffsize):
    suma = 0  # checksum256(0)
    ptr = buffer[0]
    last = buffer[-1]
    while (ptr <= last):
        ptr += 1
        suma += ptr
    return suma


def binary(e): return "{0:8b}".format(e).replace(" ", "0")


lista = [1 << i for i in range(8)]
binarios = list(map(binary, lista))
bin_diccionario = dict(zip(lista, binarios))


"""
Satellite FLAGS
"""

CONSTELLATION = dict(
    GPS=0,
    GLONASS=1,
    GALILEO=2
)

CONS = {value: key for key, value in CONSTELLATION.items()}


SIGNALS = dict(
    EPHEMERIS_AVAIL=1 << 0,
    L1_TRACK=1 << 1,
    L2_TRACK=1 << 2,
    L3_TRACK=1 << 3,
    RES4=1 << 4,
    RES5=1 << 5,
    RES6=1 << 6,
    RES7=1 << 7
)

SIGNAL_SAT = {value: key for key, value in SIGNALS.items()}


def signal_check(value):
    signals = list(filter(lambda e: e[1] > 0,
                          [(signal, nro & value) for signal, nro in
                           SIGNALS.items()]))
    return [name for name, value in signals]



import socket
from asyncio import wait_for
import asyncio
import errno
from basic_logtools.filelog import LogFile

from networktools.time import now
from networktools.library import my_random_string

from networktools.colorprint import rprint, bprint
from networktools.path import home_path


class BaseProtocol:
    """ Class to connect to tcp port and parse GSOF messages """
    tipo = "Base"

    def __init__(self, **kwargs):
        station = kwargs.get('code')
        self.sock = kwargs.get('sock')
        self.timeout = kwargs.get('timeout', 1)
        self.raise_timeout = kwargs.get('raise_timeout', False)
        self.raise_incompleteread = kwargs.get("raise_incomplete_read", False)
        self.max_try = kwargs.get("max_try", 3600)
        self.host = kwargs.get('host')
        self.port = kwargs.get('port')
        self.sock = self.create_socket(None)
        self.loop = kwargs.get('loop')
        self.station = station
        self.sock.settimeout(self.timeout)
        self.msg_dict = {}
        self.checksum = None
        self.try_connect = 0
        self.id_columns = {}
        self.keys = []
        self.status = False
        # manage asyncronous clients
        self.idc = []
        self.clients = {}
        log_path = home_path(
            kwargs.get('log_path', '~/gn_socket/logs')
        )
        log_level = kwargs.get('log_level', 'INFO')
        self.logger = LogFile(self.class_name,
                              station,
                              self.host,
                              path=log_path,
                              base_level=log_level)
        # manage asyncronous clients
        self.logger.info("Log para %s" % self.station)

    @property
    def class_name(self):
        return self.__class__.__name__

    def create_socket(self, sock):
        sock = None
        if sock is None:
            sock = socket.socket(
                socket.AF_INET, socket.SOCK_STREAM)
        else:
            sock = sock
        return sock

    def off_blocking(self):
        self.sock.setblocking(False)

    def on_blocking(self):
        self.sock.setblocking(True)

    async def connect(self):
        host = self.host
        port = self.port
        loop = self.loop
        counter = 0
        while not self.status:
            if counter > self.max_try:
                counter = 0
            try:
                future_open_conn = asyncio.open_connection(
                    loop=loop,
                    host=host,
                    port=port)
                (reader, writer) = await asyncio.wait_for(
                    future_open_conn, timeout=self.timeout)
                idc = await self.set_reader_writer(reader, writer)
                hb = await self.heart_beat(idc)
                if hb:
                    self.status = True
                    self.logger.info("Conexion a %s realizada" % self.station)
                else:
                    self.logger.error(hb)
            except asyncio.TimeoutError as te:
                self.logger.exception(
                    f"Tiempo fuera en intento de conexión {te}, iteracion {counter}")
                counter += 1
                await asyncio.sleep(10+counter)
                continue
            except socket.timeout as timeout:
                self.on_blocking()
                self.logger.error(
                    f"Error en conexión con {host}:{port}, error: {timeout}, iteración {counter}")
                # print("Error de socket a GSOF en conexión %s" % timeout)
                counter += 1
                self.status = False
                await asyncio.sleep(10+counter)
                continue
                # await self.connect()
                # raise timeout
            except socket.error as e:
                self.status = False
                counter += 1
                self.on_blocking()
                if e.errno == errno.ECONNREFUSED:
                    # print("Conexión rechazada en %s" %
                    #      self.station, file=sys.stderr)
                    self.logger.exception(
                        f"Error al conectar {e}, station {self.station}, iteración {counter}")
                    counter += 1
                    await asyncio.sleep(10+counter)
                else:
                    # print("No se puede establecer conexión, de %s" %
                    #      self.station, file=sys.stderr)
                    self.logger.exception(
                        f"Otro tipo de error al conectar {e}, station {self.station}, iteración {counter}")
                    counter += 1
                    await asyncio.sleep(10+counter)
                continue
            except (ConnectionResetError, ConnectionAbortedError) as conn_error:
                self.logger.exception(
                    f"Excepción por desconexión {conn_error}")
                continue
            except Exception as e:
                msg = f"Excepción no considerada e"
                print(msg)
                self.logger.exception(msg)
                continue
            await asyncio.sleep(10+counter)

        return idc

    # callback for create server:

    async def heart_beat(self, idc):
        tnow = now()
        if idc in self.clients.keys():
            reader = self.clients[idc]['reader']
            writer = self.clients[idc]['writer']
            closing = writer.is_closing()
            extra_info = writer.get_extra_info('peername')
            at_eof = reader.at_eof()
            if not closing and extra_info and not at_eof:
                return True
            else:
                msg_error = f"Closing {closing}, extra_info {extra_info}, at_eof {at_eof}"
                print(msg_error)
                self.logger.error(msg_error)
                self.logger.error(f"no heart_beat, at {tnow}")
                try:
                    await wait_for(self.close(idc), timeout=10)
                except asyncio.TimeoutError as te:
                    self.logger.exception(
                        f"Tiempo fuera en intento de cerrar conexión {te}, station {self.station}")
                    await asyncio.sleep(10)
                except asyncio.IncompletereadError as e:
                    self.logger.exception(
                        f"Excepción por lectura incomplete  {e}, station {self.station}")
                    await asyncio.sleep(10)
                except asyncio.CanceledError as e:
                    self.logger.exception(
                        "Excepción por error de cancelación  {e}, station {self.station}")
                    await asyncio.sleep(10)
                except (ConnectionResetError, ConnectionAbortedError) as conn_error:
                    self.logger.exception(
                        "Excepción por desconexión {conn_error}, station {self.station}")
                    await asyncio.sleep(10)
                except Exception as e:
                    print(f"Excepción no considerada, {e}")
                    self.logger.exception(
                        f"Excepción no considerada {e}, station {self.station}")
                    await asyncio.sleep(10)
                self.logger.exception(
                    "Cerrando correctamente la conexión, por no heartbeat")
                self.status = False
                return False
        else:
            self.logger.error(
                "no heart_beat, idc %s not client, code station %s" % (idc, self.station))
            await self.close(idc)
            self.status = False
            return False

    def info(self, idc):
        if self.clients:
            writer = self.clients[idc]['writer']
            return idc, writer.get_extra_info('peername')
        else:
            return idc, None

    def set_idc(self):
        """
        Defines a new id for relation process-collect_task, check if exists
        """
        uin = 4
        idc = my_random_string(uin)
        while True:
            if idc not in self.idc:
                self.idc.append(idc)
                break
            else:
                idc = my_random_string(uin)
        return idc

    async def set_reader_writer(self, reader, writer):
        idc = self.set_idc()
        # self.log_info_client(writer)
        self.clients.update(
            {idc: {
                'reader': reader,
                'writer': writer
            }
            }
        )
        return idc

    def list_clients(self):
        for i in range(len(self.conss)):
            print(str(self.addrs[i]) + ":" + str(self.conns[i]))

    async def close(self, idc):
        self.logger.error("La conexión se cerró en cliente %s" % idc)
        self.status = False
        reader = self.clients[idc]['reader']
        writer = self.clients[idc]['writer']
        writer.close()
        try:
            await wait_for(writer.wait_closed(), timeout=10)
        except asyncio.TimeoutError as te:
            print(f"Close->Error Timeout al leer {te} bytes en {self.station}")
            self.logger.exception(
                f"Station {self.station}, Tiempo fuera al leer en readbytes, {te}")
        except asyncio.IncompleteReadError as ir:
            print("Close_>Incomplete read, station {self.station}")
            self.logger.exception(
                f"Station {self.station}, Tiempo fuera al no poder leer en close, {ir}")
        except (ConnectionResetError, ConnectionAbortedError) as conn_error:
            self.logger.exception(
                f"Close->Excepción por desconexión al intentar leer {conn_error}")
        except Exception as e:
            print("Excepción no considerada")
            self.logger.exception(
                f"Close->Excepción no considerada al intentar leer e")

    async def stop(self):
        for idc in self.clients:
            await self.close(idc)

    async def readbytes(self, reader, n):
        future = reader.readexactly(n)
        try:
            timeout = self.timeout
            if n > 1:
                timeout = timeout*n
            result = await wait_for(future, timeout=timeout)
            return result
        except BrokenPipeError as be:
            print("readbytes", f"Error broken pip {be}")
            return b''
        except asyncio.IncompleteReadError as ir:
            print(
                f"Error por lectura incomplete al leer {n} bytes en {self.station}")
            self.logger.exception(
                f"Station {self.station}, lectura incomplete al leer en readbytes, error: {ir}")
        except asyncio.TimeoutError as te:
            print(f"Error Timeout al leer {n} bytes en {self.station}")
            self.logger.exception(
                f"Station {self.station}, Tiempo fuera al leer en readbytes, {te}")
            if self.raise_timeout:
                raise te
            else:
                return b''
        except asyncio.IncompleteReadError as ir:
            print(f"Incomplete read, station {self.station}")
            self.logger.exception(
                f"Station {self.station}, Tiempo fuera al no poder leer en readbytes {n} bytes, {ir}")
            return b''
        except (ConnectionResetError, ConnectionAbortedError) as conn_error:
            self.logger.exception(
                f"Excepción por desconexión al intentar leer {conn_error}")
            return b''
        except Exception as e:
            print("Excepción no considerada")
            self.logger.exception(
                "Excepción no considerada al intentar leer {e}")
            return b''

    async def get_message_header(self, idc):
        self.idc = idc

    async def get_records(self):
        """
        Esta corutina toma los datos en bytes y los transforma,
        debe entregar una verificaciond e checksum y un diccionario
        """
        return False, {}



from dataprotocols.gsof import Gsof
from dataprotocols.eryo import Eryo
from pathlib import Path
import asyncio
import functools
import os
import signal
import click
import json
import time
from networktools.ip import validURL


PROTOCOL = dict(GSOF=Gsof, ERYO=Eryo)


async def run_test(loop, code, host, port, limit=2000, show=True,
                   protocol=Gsof, data_file='./test.json'):
    ginput = dict(host=host,
                  port=port,
                  code=code,
                  timeout=5,
                  sock=None,
                  loop=loop)
    protocol_name = protocol.__class__.__name__
    print("Inicializando %s con:" % (protocol_name))
    [print(k, "->", v) for k, v in ginput.items()]
    file_path = Path(data_file)
    limit_counter = limit
    counter = 0
    print("Conexion a->", ginput)
    instance = protocol(**ginput)
    idc = await instance.connect()
    print("IDC->", idc)
    with open(file_path, "a+") as prt_file:
        prt_file.write("[")
        while counter <= limit_counter:
            try:
                print("Heart beat")
                heartbeat = await instance.heart_beat(idc)
                print("Heartbeat", heartbeat)
                if heartbeat:
                    print("MSG recv>")
                    await instance.get_message_header(idc)
                    done, msg = await instance.get_records()
                    if done:
                        if show:
                            print("%s -> %s" % (protocol_name, code))
                            [print(k, "->", v) for k, v in msg.items()]
                        json.dump(msg, prt_file, indent=2)
                        prt_file.write(",\n")
                else:
                    await instance.close(idc)
                    idc = await instance.connect()
            except Exception as ex:
                prt_file.write("]")
                await instance.close(idc)
                print(f"Error {ex}")
                loop.close()
                raise ex
            except KeyboardInterrupt as ke:
                prt_file.write("]")
                await instance.close(idc)
                loop.close()
                print(ke)
                raise ke
            counter += 1
        prt_file.write("]")
    await instance.close(idc)


@click.command()
@click.option("--code", default="CODE", help="Código de estación (obligatoria)")
@click.option("--host", default="AQUI-VA-LA-URL", help="URL de estación (obligatoria)")
@click.option("--port", default=9080, help="Nro de puerto, defecto 9080")
@click.option("--limit", default=2000, help="Nro de mensajes, defecto 2000")
@click.option("--show/--no-show", default=True, show_default=True,
              type=bool, help="Mostrar cada mensaje")
@click.option("--protocol", default="GSOF", help="Protocolo, defecto GSOF, opciones {GSOF, ERYO}")
@click.option("--data_file", default='./test.json', help="Archivo de almacenamiento de datos, defecto ./test.json")
def run_protocol(code, host, port, limit, show, protocol, data_file):
    loop = asyncio.get_event_loop()
    try:
        if validURL(host):
            prt = PROTOCOL.get(protocol.upper(), Gsof)
            loop.run_until_complete(
                run_test(loop, code, host, port, limit, show, prt, data_file))
        else:
            print("Ingresa datos reales para obtener resultados: protocol --help\nRevisa tus parámetros nuevamente")
    except Exception as ex:
        print("Exception %s" % ex)
        loop.call_soon_threadsafe(loop.stop)
        print(ex)
    except KeyboardInterrupt as ke:
        print("Exception %s" % ke)
        loop.call_soon_threadsafe(loop.stop)
        print(ke)
    finally:
        loop.close()


if __name__ == '__main__':
    run_protocol()



from networktools.time import timestamp
gsof_table_names = {
    1: "TIME",
    2: "LATLONHEIGHT",
    3: "ECEF",
    4: "Local Datum Position".upper(),
    5: "Local_Zone_Position".upper(),
    6: "ECEF_DELTA",
    7: "TPlane_ENU".upper(),
    8: "Velocity".upper(),
    9: "DOP",
    10: "Clock_Info".upper(),
    11: "Position_VCV".upper(),
    12: "SIGMA",
    13: "GPS_SV_Brief".upper(),
    14: "GPS_SV_Detail".upper(),
    15: "Receiver_Serial_Number".upper(),
    16: "UTC",
    27: "Attitude".upper(),
    33: "All_SV_Brief".upper(),
    48: "All_SV_Detail".upper(),
    # http://www.trimble.com/OEM_ReceiverHelp/v4.85/en/GSOFmessages_ReceivedBaseInfo.html
    35: "Received_Base_Info".upper(),
    37: "Batt_Mem".upper(),
    40: "L-Band".upper(),
    41: "BASE_POSITION_AND_QUALITY_INDICATOR",
}
gsof_field_names = {
    1: ('GPS_TIME', 'GPS_WEEK', 'SVN_NUM', 'FLAG_1', 'FLAG_2', 'INIT_NUM'),
    2: ('LATITUDE', 'LONGITUDE', 'HEIGHT'),
    3: ('X_POS', 'Y_POS', 'Z_POS'),
    4: ('LOCAL_DATUM_ID', 'LOCAL_DATUM_LAT', 'LOCAL_DATUM_LON', 'LOCAL_DATUM_HEIGHT', 'OPRT'),
    5: ('LOCAL_DATUM_ID', 'LOCAL_ZONE_ID', 'LOCAL_ZONE_NORTH', 'LOCAL_ZONE_EAST', 'LOCAL_DATUM_HEIGHT'),
    6: ('DELTA_X', 'DELTA_Y', 'DELTA_Z'),
    7: ('DELTA_EAST', 'DELTA_NORTH', 'DELTA_UP'),
    8: ('VEL_FLAG', 'VELOCITY', 'HEADING', 'VERT_VELOCITY'),
    9: ('PDOP', 'HDOP', 'VDOP', 'TDOP'),
    10: ('CLOCK_FLAG', 'CLOCK_OFFSET', 'FREQ_OFFSET'),
    # http://www.trimble.com/OEM_ReceiverHelp/v4.85/en/GSOFmessages_PositionVCV.html
    11: ('POSITION_RMS_VCV', 'VCV_XX', 'VCV_XY', 'VCV_XZ', 'VCV_YY', 'VCV_YZ', 'VCV_ZZ', 'UNIT_VAR_VCV', 'NUM_EPOCHS_VCV'),
    12: ('POSITION_RMS_SIG', 'SIG_EAST', 'SIG_NORT', 'COVAR_EN', 'SIG_UP', 'SEMI_MAJOR', 'SEMI_MINOR', 'ORIENTATION', 'UNIT_VAR_SIG', 'NUM_EPOCHS_SIG'),
    13: ('NUM_SV', 'PRN', 'SV_FG1', 'SV_FG2'),
    14: ('PRN', 'FLAGS1', 'FLAGS2', 'ELEVATION', 'AZIMUTH', 'SNR_L1', 'SNR_L2'),
    15: 'SERIAL_NUM',
    16: ('GPS_MS_OF_WEEK', 'GPS_WEEK', 'UTC_OFFSET', 'CT_FLAGS'),
    27: ('GPS_TIME', 'FLAGS', 'N_SV', 'CALC_MODE', '', 'PITCH', 'YAW', 'ROLL',
         'MSRANGE', 'PDOP', 'PITCHV', 'YAWV', 'ROLLV', 'PITCH_YAW_COVAR',
         'PITCH_ROLL_COVAR', 'YAW_ROLL_COVAR', 'MSRANGE_VAR'),
    33: ('NUM_SV', 'SVSYS', 'SVFLAG_1', 'SVFLAG_2'),
    48: ('PRN', 'SV_SYSTEM', 'SV_FLAG1', 'SV_FLAG2', 'ELEVATION', 'AZIMUTH', 'SNR_FIRST', 'SNR_SECOND', 'SNR_THIRD'),
    35: ('FLAGS', 'BASE_NAME', 'BASE_ID', 'BASE_LAT', 'BASE_LON', 'BASE_HEIGHT'),
    37: ('BATT_CAPACITY', 'REMAINING_MEM'),
    40: ('SAT_NAME', 'SAT_FREQ', 'SAT_BITRATE', 'SNR', 'HP/XP_MODE', 'VBS_MODE',
         'BEAM_MODE', 'O_MOTION', '3SHTHRES', '3SVTHRES', 'NMEA', 'IQ_RATIO',
         'EST_BER', 'TOT_UNIQ_WORDS', 'TOT_BAD_U_W_BITS', 'TOT_VITERBI', 'TOT_CORRECT_VITERBI',
         'TOT_BAD_MSG', 'MEAS_F_FLAG', 'MEAS_F'),
    41: ('GPS_TIME', 'GPS_WEEK_NR', 'LATITUDE', 'LONGITUDE', 'HEIGHT', 'QI')
}

# print(rec_field_names)
"""
   C source code:
    https://www.trimble.com/OEM_ReceiverHelp/v5.11/en/AppNote_GSOFMsging.html#LBand

 """
gsof_rec = {
    1: '>LH4B',
    2: '>3d',
    3: '>3d',
    4: '>8s3dB',
    5: '>2s3d',
    6: '>3d',
    7: '>3d',
    8: '>B3f',
    9: '>4f',
    10: '>B2d',
    11: '>8fh',
    12: '>9fh',
    13: ('>B', '>3B'),
    # 13 *number of sv
    14: ('>B','>4BH2B'),
    # http://www.trimble.com/OEM_ReceiverHelp/v4.85/en/Default.html#GSOFmessages_SVDetail.html
    15: '>l',
    16: '>l2hB',
    27: '>L4B4dH7f',
    33: ('>B', '>4B'),
    # http://www.trimble.com/OEM_ReceiverHelp/v4.85/en/Default.html#GSOFmessages_AllSVBrief.html
    # 33 *number of sv
    48: ('>B', '>5BB3B'),##BIG ???
    #http://www.trimble.com/OEM_ReceiverHelp/v4.85/en/Default.html#GSOFmessages_AllSVDetail.html
    35: '>B8B2B3d',
    37: '>Hd',
    40: '>5BfHf5B2fB2f6LBd',
    # INCOMPLETE: http://www.trimble.com/OEM_ReceiverHelp/v4.85/en/Default.html#GSOFmessages_LBand.html
    41: '>lh3dB'
}

from networktools.time import timestamp

def logfile(station):
    return './logs/gsof_%s_%s.log' % (station, str(timestamp())[:-7] )





from setuptools import setup, find_packages
from pathlib import Path

path = Path(__file__).resolve().parent
with open(path/'README.md', encoding='utf-8') as f:
    long_description = f.read()

with open(path/'VERSION') as version_file:
    version = version_file.read().strip()

setup(name='dataprotocols',
      version=version,
      description='DataProtocols is a set of classes that implement protocols for data acquisition',
      url='http://gitlab.csn.uchile.cl/dpineda/dataprotocols',
      author='David Pineda Osorio',
      author_email='dpineda@csn.uchile.cl',
      install_requires=['Click', 'networktools', 'basic-logtools'],
      scripts=[
          'dataprotocols/scripts/gsof.py',
          'dataprotocols/scripts/eryo.py',
          'dataprotocols/scripts/protocol.py',
      ],
      entry_points={
          'console_scripts': ["gsof = dataprotocols.scripts.gsof:run_gsof",
                              "eryo = dataprotocols.scripts.eryo:run_eryo",
                              "protocol = dataprotocols.scripts.protocol:run_protocol"]
      },
      packages=find_packages(),
      include_package_data=True,
      license='GPLv3',
      long_description=long_description,
      long_description_content_type='text/markdown',
      zip_safe=False)
